//
// Messaging.h
//
// this file contains the framework that provides
// the Java-style messaging system
// to the C++ objects
//
// Copyright (C) 2001 Maciej Sobczak
//
// you can use this code for any purpose without limitations
// (and for your own risk) as long as this notice remains
//

#ifndef INCLUDED_MESSAGING_H
#define INCLUDED_MESSAGING_H

#include <vector>
#include <algorithm>

// this class serves as an empty fill-in for synchronization
// purposes you can provide your own class but it should have
// these two methods, as well as working and accessible
// default/copy constructor and operator=

class NoSynchronization
{
public:
	void lock() {}
	void unlock() {}
};

// this class is a bag for pointers
// but with synchronization added
//
// uses object of type Mutex for synchronizing
// the requirements for a Mutex class are:
// 1. lock() and unlock() methods should be present
// 
// contains the collection object to store
// pointers, it is also possible to instantiate
// this class with different back-end collection classes
// the requirements for the BagType class are:
// 1. push_back() method
// 3. end() method
// 4. erase() method (that accepts iterator)
// 5. iterator type defined (InputIterator)
// in addition, this class should be Assignable
template
<
	class Listener,
	class Mutex,
	class BagType
>
class ListenersCollection
{
	// this is a helper class that locks a given mutex
	// and unlocks when exiting the scope
	// (it is therefore exception-aware)
	class MutexHolder
	{
	public:
		MutexHolder(Mutex &m)
			: mutex_(m)
		{
			mutex_.lock();
		}
		~MutexHolder()
		{
			mutex_.unlock();
		}
	private:
		Mutex &mutex_;
	};

public:

	ListenersCollection(Mutex &m) : mutex_(m) {}

	void add(Listener *p)
	{
		MutexHolder mh(mutex_);
		bag_.push_back(p);
	}
	void remove(Listener *p)
	{
		typedef typename BagType::iterator iterator;

		MutexHolder mh(mutex_);
		iterator it =
			std::find(bag_.begin(), bag_.end(), p);
		if (it == bag_.end())
			return;
		else
			bag_.erase(it);
	}

	BagType getFrozen() const
	{
		MutexHolder mh(mutex_);
		return bag_;
	}

private:

	BagType bag_;
	Mutex &mutex_;
};

// this class serves as a fundamental building block for
// a Java-style messaging system
//
// Listener is a type of the interface that can be registered as a listener
// Context is a type of the context that gives a clue while dispatching the event
// throught the pointer to the listener interface
// Mutex and BagType are explained above
template
<
	class Listener,
	typename Context = int,
	class Mutex = NoSynchronization,
	class BagType = std::vector<Listener*>
>
class Messaging
{
public:
	Messaging(Mutex m = Mutex()) : listeners_(m) {}

	// registers the listener
	// listeners registered more than once will be notified once for
	// each registration
	void addListener(Listener *listener)
	{
		listeners_.add(listener);
	}

	// unregisters the listener
	// it doesn't hurt if it's not already registered
	// multiply registered listeners should also multiply unregister themselves
	void removeListener(Listener *listener)
	{
		listeners_.remove(listener);
	}

protected:

	// this method should be called by the derived class to trigger
	// event dispatching
	void raiseEvent(const Context &context)
	{
		typedef typename BagType::iterator iterator;

		BagType frozen(listeners_.getFrozen());
		iterator it;
		iterator itend = frozen.end();
		Listener *p;

		for (it = frozen.begin(); it != itend; ++it)
		{
			p = static_cast<Listener*>(*it);
			dispatchEvent(p, context);
		}
	}

	// this is intended to be overriden in a derived class
	// and is responsible for the actual dispatching
	// this method is called for every registered listener (once for each registration)
	// the overriding method should call the listener, using (or not) the context
	// as a clue
	virtual void dispatchEvent(Listener *listener, const Context &context) = 0;

private:

	// the synchronized container for keeping pointers to listener objects
	ListenersCollection<Listener, Mutex, BagType> listeners_;
};

#endif
