/*
	Copyright (c) 2003 Michael Carrato. All rights reserved.	

	Permission to use, modify, copy, and distribute this software 
	and its documentation is hereby granted without fee, 
	provided that the above copyright notice appears in all copies.
	This software is provided "as is" without express or implied 
    warranty.
*/

//disable annoying MSC warning about debug symbols
#ifdef _MSC_VER
#pragma warning (disable:4786)
#endif // _MSC_VER

#include "CompactMap.h"
#include <map>

#include <iostream>

class TestMapKey
{
public:
    TestMapKey(int v)
        : val(v)
    {
    }
    bool operator==(const TestMapKey& rhs) const
    {
        return val == rhs.val;
    }
    bool operator!=(const TestMapKey& rhs) const
    {
        return !operator==(rhs);
    }
    bool greaterThan(const TestMapKey& rhs) const
    {
        return val > rhs.val;
    }
private:
    int val;
};

struct TestMapKeyLess
: public std::binary_function<TestMapKey,TestMapKey,bool>
{
    bool operator()(const TestMapKey& lhs,const TestMapKey& rhs) const
    {
        return lhs.greaterThan(rhs);
    }
};

void testSEMap()
{
    typedef CompactMap<TestMapKey,int,TestMapKeyLess,MinimizeSpacePolicy> CompactMapType;
    typedef std::map<TestMapKey,int,TestMapKeyLess> ControlMapType;
//    typedef CompactSet<int,std::less<int>,MinimizeSpacePolicy> CompactMapType;
    ControlMapType control;
    CompactMapType sa;
    for(int i=0;i<500000;++i)
    {
	if(i % 1000 == 0)
	{
		std::cout << i << ", size=" << sa.size() << "              \r" << std::flush;
	}
        unsigned action = (rand() % 7001)/1000;
        if(action == 7)
            action += (rand() % 2);
        switch(action)
        {
        case 0:
            {
				//insert
                unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                unsigned newRand2 = (unsigned) (rand() * rand()) % 100000;
                if(rand() % 2 == 0)
                {
                    //use insert()
                    std::pair<ControlMapType::iterator,bool> cRet = control.insert(std::make_pair(newRand,newRand2));
                    std::pair<CompactMapType::iterator,bool> csRet = sa.insert(std::make_pair(newRand,newRand2));
                    assert(cRet.second == csRet.second);
                    assert(*cRet.first == *csRet.first);
                    --(++csRet.first);
                    assert(*cRet.first == *csRet.first);
                }
                else
                {
                    assert(control[newRand] == sa[newRand]);
                    control[newRand] = newRand2;
                    sa[newRand] = newRand2;
                    assert(control[newRand] == sa[newRand]);
                }
            }
            break;
        case 1:
            {
				//erase value
				unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                ControlMapType::size_type cRet = control.erase(newRand);
                CompactMapType::size_type csRet = sa.erase(newRand);
                assert(cRet == csRet);
            }
            break;
        case 2:
            {
				//find
                unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                ControlMapType::iterator cRet = control.find(newRand);
                CompactMapType::iterator csRet = sa.find(newRand);
                if(cRet == control.end())
                {
                    assert(csRet == sa.end());
                }
                else
                {
					assert(*csRet == *cRet);
                    --(++csRet);
					assert(*csRet == *cRet);
                }
            }
            break;
        case 3:
            {
				//constant find
                unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                const ControlMapType& constControl = control;
				const CompactMapType& constSA = sa;
				ControlMapType::const_iterator cRet = constControl.find(newRand);
                CompactMapType::const_iterator csRet = constSA.find(newRand);
//				CompactMapType::iterator csRetFAIL = constSA.find(newRand); //SHOULD FAIL
                if(cRet == constControl.end())
                {
                    assert(csRet == constSA.end());
                }
                else
                {
					assert(*csRet == *cRet);
                    --(++csRet);
					assert(*csRet == *cRet);
                }
            }
            break;
		case 4:
#if ! defined(_MSC_VER) || _MSC_VER > 1200
			{
				//insert bulk
				std::vector<std::pair<TestMapKey,int> > vals;
				for(int insCount = rand() % 1000;insCount; --insCount)
				{
					unsigned newRand = (unsigned) (rand() * rand()) % 100000;
					unsigned newRand2 = (unsigned) (rand() * rand()) % 100000;
					vals.push_back(std::make_pair(newRand,newRand2));
				}
				control.insert(vals.begin(),vals.end());
				sa.insert(vals.begin(),vals.end());
			}
#endif
			break;
		case 5:
			{
				//count
                unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                ControlMapType::size_type ccontrol = control.count(newRand);
                CompactMapType::size_type csa = sa.count(newRand);
				assert(ccontrol == csa);
			}
			break;
		case 6:
			{
				//bounds functions
                unsigned newRand = (unsigned) (rand() * rand()) % 100000;
                ControlMapType::iterator controlLB = control.lower_bound(newRand);
                CompactMapType::iterator cmLB = sa.lower_bound(newRand);
                if(controlLB == control.end())
                {
                    assert(cmLB == sa.end());
                }
                else
                {
					assert(*controlLB == *cmLB);
                    --(++cmLB);
					assert(*controlLB == *cmLB);
                }
                ControlMapType::iterator controlUB = control.upper_bound(newRand);
                CompactMapType::iterator cmUB = sa.upper_bound(newRand);
                if(controlUB == control.end())
                {
                    assert(cmUB == sa.end());
                }
                else
                {
					assert(*controlUB == *cmUB);
                    --(++cmLB);
					assert(*controlUB == *cmUB);
                }
				std::pair<CompactMapType::iterator,CompactMapType::iterator>
					er = sa.equal_range(newRand);
				assert(er.first == cmLB);
				assert(er.second == cmUB);
			}
			break;
		case 7:
			{
				//erase iterator
				assert(control.size() == sa.size());
				if(!control.size())
					break;
				int i,count = (rand() * rand()) % (control.size());
				ControlMapType::iterator controlIt = control.begin();
				CompactMapType::iterator compactIt = sa.begin();
				for(i=0; i != count; ++i)
				{
					++controlIt;
					++compactIt;
				}

                assert(compactIt->first == controlIt->first);
                assert(compactIt->second == controlIt->second);

				control.erase(controlIt++);
				compactIt = sa.erase(compactIt);
                if(controlIt == control.end())
                {
                    assert(compactIt == sa.end());
                }
                else
                {
                    assert(*compactIt == *controlIt);
                    --(++compactIt);
                    assert(*compactIt == *controlIt);
                }
			}
			break;
        }
        if(rand() % 200 == 0)
        {
            //test copy constructor, assignment, and swap
            CompactMapType copy(sa);
            assert(sa == copy);
            if(sa.size())
            {
                sa.erase(sa.begin());
                assert(sa != copy);
            }
            std::swap(sa,copy);
            assert(sa.size() == control.size());
            std::swap(sa,copy);
            sa = copy;

            //test forward and reverse iterators (also validates that the
            //above tests succeeded)
            std::vector<std::pair<TestMapKey,int> > vec1,vec2,vec3;
	        std::copy(sa.begin(),sa.end(),std::back_inserter(vec1));
	        std::copy(sa.rbegin(),sa.rend(),std::back_inserter(vec3));
	        std::reverse(vec3.begin(),vec3.end());
	        std::copy(control.begin(),control.end(),std::back_inserter(vec2));
	        assert(vec1 == vec2);
	        assert(vec3 == vec2);
        }
        if(rand() % 10000 == 0)
        {
            sa.clear();
            control.clear();
        }
/*        std::copy(control.begin(),control.end(), std::ostream_iterator<unsigned>(std::cout,","));
        std::cout << "\n";
        std::copy(sa.begin(),sa.end(), std::ostream_iterator<int>(std::cout,","));
        std::cout << "\n\n";*/
    }
    sa.compact();
    std::cout << "total elements: test=" << sa.size() 
            << ", control=" << control.size() << "\n";
    ControlMapType::const_iterator ctrlIt, ctrlItEnd = control.end();
    CompactMapType::const_iterator saIt, saItEnd = sa.end();
    for(ctrlIt = control.begin(), saIt = sa.begin(); 
            ctrlIt != ctrlItEnd && saIt != saItEnd;
            ++ ctrlIt, ++saIt)
    {
        if(ctrlIt->first != saIt->first || ctrlIt->second != saIt->second)
            break;
    }
    if(ctrlIt == ctrlItEnd && saIt == saItEnd)
    {
        std::cout << "test identical to control\n";
    }
    else
    {
        std::cout << "ERROR: test different from control!\n";
    }
//    char c;
//    std::cin >> c;
}
