/*
	Copyright (c) 2003 Michael Carrato. All rights reserved.	

	Permission to use, modify, copy, and distribute this software 
	and its documentation is hereby granted without fee, 
	provided that the above copyright notice appears in all copies.
	This software is provided "as is" without express or implied 
    warranty.
*/

#if !defined(AFX_COMPACTMAP_H__BC32B669_4C12_4E9A_949B_833EFEBE2493__INCLUDED_)
#define AFX_COMPACTMAP_H__BC32B669_4C12_4E9A_949B_833EFEBE2493__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


#include "CompactColl.h"
#include <map>

//compact map collection traits
template <class K, class V, class P,class POL, class A>
struct CMapCT
{
    typedef K key_type;
    typedef V mapped_type;
    typedef std::pair<const K,V> value_type;
    typedef P key_compare;
    struct value_compare
        : public key_compare
    {
        value_compare()
        { }
        value_compare(const key_compare& rhs)
            : key_compare(rhs)
        { }
        bool operator()(const value_type& lhs,const value_type& rhs) const
        {
            return key_compare::operator()(lhs.first,rhs.first);
        }
        bool operator()(const key_type& lhs,const key_type& rhs) const
        {
            return key_compare::operator()(lhs,rhs);
        }
        bool operator()(const key_type& lhs,const value_type& rhs) const
        {
            return key_compare::operator()(lhs,rhs.first);
        }
        bool operator()(const value_type& lhs,const key_type& rhs) const
        {
            return key_compare::operator()(lhs.first,rhs);
        }
    };

	class AssignableValueType
		: public value_type
	{
	public:
		AssignableValueType(const key_type& k, const mapped_type& v)
			: value_type(k,v)
		{
		}
		AssignableValueType(const value_type& rhs)
			: value_type(rhs)
		{
		}
		AssignableValueType& operator=(const value_type& rhs)
		{
			const_cast<key_type&>(first) = rhs.first;
			second = rhs.second;
			return *this;
		}
		AssignableValueType& operator=(const AssignableValueType& rhs)
		{
			const_cast<key_type&>(first) = rhs.first;
			second = rhs.second;
			return *this;
		}
	};
			
    typedef A allocator_type;
    typedef typename allocator_type::size_type size_type;
    typedef POL PolicyTagType;
#if defined(_MSC_VER) && _MSC_VER <= 1200
	typedef std::allocator<AssignableValueType> AssignableAllocatorType;
#else
	typedef typename allocator_type::template rebind<AssignableValueType>::other  AssignableAllocatorType;
#endif
    typedef std::map<K,V,P,A> DynSetType;

    value_compare kc;
    allocator_type al;
    CMapCT(const key_compare& pred, 
        const allocator_type& alloc)
        : kc(pred), al(alloc)
    { }
    CMapCT()
    { }

    void swap(CMapCT& rhs)
    {
        std::swap(kc,rhs.kc);
        std::swap(al,rhs.al);
    }
    bool less(const key_type& lhs, const key_type& rhs) const
    {
        return kc(lhs,rhs);
    }
    void assign(value_type& lhs, const value_type& rhs) const
    {
		//need to do a const cast to do this assignment... to get
		//rid of the const key
		const_cast<key_type&>(lhs.first) = rhs.first;
		lhs.second = rhs.second;
	}
    const key_type& extractKey(const value_type& val) const
    {
        return val.first;
    }
    const key_type& extractKey(value_type& val) const
    {
        return val.first;
    }
    const key_compare& key_comp() const
    {
        return kc;
    }
    const value_compare& value_comp() const
    {
        return kc;
    }
    const allocator_type& get_allocator() const
    {
        return al;
    }
};

#ifdef _MSC_VER
namespace std
{
template<class T>
struct MSCStdSwapInjector
{
    friend void swap(T& lhs,T& rhs)
    {
        lhs.swap(rhs);
    }
};
} //namespace std
#endif

template<class K,class V,class P = std::less<K>,class POLICY = MaximizeSpeedPolicy,class A = std::allocator<std::pair<const K,V> > >
class CompactMap
: public CompactColl<CMapCT<K,V,P,POLICY,A> >
#ifdef _MSC_VER
, public std::MSCStdSwapInjector<CompactMap>
#endif
{
	typedef CMapCT<K,V,P,POLICY,A> CTType;
    typedef CompactColl<CTType> ImplType;
public:
    typedef typename ImplType::key_compare key_compare;
    typedef typename ImplType::allocator_type allocator_type;
    typedef typename ImplType::key_type key_type;
    typedef V mapped_type;

    explicit CompactMap(const key_compare& pred = key_compare(), 
        const allocator_type& alloc = allocator_type())
        : ImplType(CMapCT<K,V,P,POLICY,A>(pred,alloc))
    {
    }
    mapped_type& operator[](const key_type& key)
    {
        return insert(value_type(key,mapped_type())).first->second;
    }
};

template<class K,class V,class P,class POLICY,class A>
inline bool operator==(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return (lhs.size() == rhs.size() 
        && std::equal(lhs.begin(),lhs.end(),rhs.begin()));
}

template<class K,class V,class P,class POLICY,class A>
inline bool operator!=(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return ! (lhs == rhs);
}

template<class K,class V,class P,class POLICY,class A>
inline bool operator<(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return std::lexicographical_compare(lhs.begin(),lhs.end(),rhs.begin(),rhs.end());
}

template<class K,class V,class P,class POLICY,class A>
inline bool operator>(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return rhs < lhs;
}

template<class K,class V,class P,class POLICY,class A>
inline bool operator<=(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return ! (rhs < lhs);
}

template<class K,class V,class P,class POLICY,class A>
inline bool operator>=(const CompactMap<K,V,P,POLICY,A>& lhs, const CompactMap<K,V,P,POLICY,A>& rhs)
{
    return ! (lhs < rhs);
}

#ifndef _MSC_VER
namespace std
{

template<class K,class V,class P,class POLICY,class A>
inline void swap(::CompactMap<K,V,P,POLICY,A>& lhs, ::CompactMap<K,V,P,POLICY,A>& rhs)
{
    lhs.swap(rhs);
}
}
#endif

#endif // !defined(AFX_COMPACTMAP_H__BC32B669_4C12_4E9A_949B_833EFEBE2493__INCLUDED_)
